@@ -20,8 +20,9 @@
 
 import re
 import sys
+import copy
 from email.utils import getaddresses
-from authheaders.dmarc_lookup import dns_query, receiver_record, get_org_domain
+from authheaders.dmarc_lookup import dns_query, receiver_record, receiver_record_walk, get_org_domain
 from authres import SPFAuthenticationResult, DKIMAuthenticationResult, AuthenticationResultsHeader
 from authres.arc import ARCAuthenticationResult
 from authres.dmarc import DMARCAuthenticationResult
@@ -51,6 +52,216 @@ def get_domain_part(address):
     return res[0].lower().decode('ascii')
 
 
+def check_psddmarc_list(psdname, dnsfunc=dns_query):
+    """Check psddmarc.org list of PSD DMARC participants"""
+    try:
+        # If the PSD registry is locally available, use it.
+        psdfile_name = resource_filename('authheaders', 'psddmarc.csv')
+        psd_file = open(psdfile_name)
+        psds = []
+        for line in psd_file.readlines():
+            sp = line.split(',')
+            if sp[1] == 'Active':
+                psds += sp[0]
+        if psdname in psds:
+            return True
+        else:
+            return False
+    except:
+        # If not, use the DNS query list.
+        psd_list_host = '.psddmarc.org'
+        psd_lookup = psdname + psd_list_host
+        answer = dnsfunc(psd_lookup)
+        if answer:
+            return True
+        else:
+            return False
+
+
+def dmarc_per_from(from_domain, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=False, dmarcbis=False, policy_only=False):
+    """DMARC result for a single From domain."""
+    original_from = from_domain
+    psddomain = False
+    if not dmarcbis: # It's all different in the future
+        # Get dmarc record for domain
+        if(dnsfunc):
+            record, orgdomain = receiver_record(from_domain, dnsfunc=dnsfunc)
+        else:
+            record, orgdomain = receiver_record(from_domain)
+        # Report if DMARC record is From Domain or Org Domain
+        if record and orgdomain:
+            result_comment = 'Used Org Domain Record'
+            policydomain = orgdomain
+        elif record:
+            result_comment = 'Used From Domain Record'
+            orgdomain = from_domain
+            policydomain = None
+
+        # Get psddmarc record if doing PSD DMARC, no DMARC record, and PSD is
+        #  listed
+        if (not record) and psddmarc:
+            org_domain = get_org_domain(from_domain)
+            if(dnsfunc):
+                if check_psddmarc_list(org_domain.split('.',1)[-1],
+                                       dnsfunc=dnsfunc):
+                    record, _ = receiver_record(org_domain.split('.',1)[-1],
+                                                dnsfunc=dnsfunc)
+            else:
+                if check_psddmarc_list(org_domain.split('.',1)[-1]):
+                    record, _ = receiver_record(org_domain.split('.',1)[-1])
+            if record:
+                psddomain = org_domain.split('.',1)[-1]
+                result_comment = 'Used Public Suffix Domain Record'
+    else:
+        # Get dmarc record for domain (tree walk)
+        # TODO: Not very efficient.  Always does tree walk, even if not
+        # really needed.  receiver_record_walk does sequential DNS lookups
+        # vice parallel.
+        orgdomain = False
+        record = None
+        if(dnsfunc):
+            treeresults = receiver_record_walk(from_domain, dnsfunc=dnsfunc)
+        else:
+            treeresults = receiver_record_walk(from_domain)
+        # Then find org domain, per DMARCbis 07
+        # Fake psd=yes (since no one publishes this yet) - FIXME later
+        for dmn, rec in reversed(list(treeresults.items())):
+            if check_psddmarc_list(dmn):
+                treeresults[dmn]['psd'] = 'y'
+                break
+        orgresults = copy.deepcopy(treeresults)
+        psddomain = False
+        for dmn, rec in list(treeresults.items()):
+            try:
+                psd = treeresults[dmn]['psd']
+                # For psd=y, use next longer domain as org.
+                if psd == 'y':
+                    if dmn == from_domain:
+                        orgdomain = dmn
+                        record = rec
+                        result_comment = 'Used From Domain which is also PSD'
+                    else:
+                        psddots = len(dmn[0].split('.'))
+                        fromsplit = from_domain.split('.')
+                        orgdomain = ''
+                        start = True
+                        for part in fromsplit[(len(fromsplit) - psddots - 1):]:
+                            if not start:
+                                orgdomain += '.'
+                            start = False
+                            orgdomain += part
+                        try:
+                            if treeresults[orgdomain]:
+                                record = treeresults[orgdomain]
+                        except:
+                            record = rec
+                            psddomain = dmn
+                        result_comment = 'Used Tree Walk, org one level below PSD'
+                    break
+                # For psd=n, this is the org domain.
+                if psd == 'n':
+                    orgdomain = dmn
+                    record = rec
+                    result_comment = 'Used Tree Walk Record which is PSD=n'
+                    break
+            except:
+                # No PSD tag
+                pass
+        if not orgdomain:
+            # If psd=u (implicit or explicit), usual case.
+            for dmn, rec in reversed(list(treeresults.items())):
+                if rec and dmn != from_domain:
+                    orgdomain = dmn
+                    record = rec
+                    result_comment = 'Used Tree Walk Record'
+                    break
+        if not orgdomain:
+            # If all else fails, the From domain is the org domain.
+            orgdomain = from_domain
+            try:
+                record = treeresults[from_domain]
+                result_comment = 'Used From Domain Record'
+            except:
+                record = False
+                result_comment = 'From domain has no DMARC record'
+
+
+    if record and record.get('p'): # DMARC P tag is mandatory
+        # find policy
+        policy = record['p']
+        if policy[-1:] == '\\':
+            policy = policy[:-1]
+        try:
+            sp = record['sp']
+            if sp[-1:] == '\\':
+                sp = sp[:-1]
+        except KeyError:
+            sp = policy
+        try:
+            np = record['np']
+            if np[-1:] == '\\':
+                np = np[:-1]
+        except KeyError:
+            np = None
+
+        if orgdomain or psddomain:
+            if np:
+                exists = False
+                for qtype in ['a', 'mx', 'aaaa']:
+                    if(dnsfunc):
+                        res = dnsfunc(from_domain)
+                    else:
+                        res = dns_query(from_domain, qtype)
+                    if res:
+                        exists = True
+                        break
+                if exists:
+                    policy = sp
+                else:
+                    policy = np
+            else:
+                policy = sp
+
+        adkim = record.get('adkim', 'r')
+        aspf  = record.get('aspf',  'r')
+
+        # get result
+        result = "fail"
+        if not policy_only and spf_result and spf_result.result == "pass":
+            # The domain in SPF results often includes the local part, even though
+            # generally it SHOULD NOT (RFC 7601, Section 2.7.2, last paragraph).
+            mail_from_domain = get_domain_part(spf_result.smtp_mailfrom)
+            spf_result.smtp_mailfrom = mail_from_domain
+            if aspf == "s" and from_domain == mail_from_domain:
+                result = "pass"
+            elif aspf == "r" and get_org_domain(from_domain) == get_org_domain(mail_from_domain):
+                result = "pass"
+
+        if not policy_only and dkim_result and dkim_result.result == "pass":
+            if adkim == "s" and from_domain == dkim_result.header_d:
+                result = "pass"
+            elif adkim == "r" and get_org_domain(from_domain) == get_org_domain(dkim_result.header_d):
+                result = "pass"
+    else:
+        # If no DMARC record, no result
+        result = 'none'
+        result_comment = ''
+        from_domain = ''
+        policy = ''
+
+    if policy_only:
+        if psddomain:
+            policydomain = psddomain
+        else:
+            policydomain = orgdomain
+        if not record:
+            result_comment = "None"
+            policydomain = None
+        results = [original_from, policydomain, result_comment, policy, record, orgdomain]
+        return(results)
+    else:
+        return(result, result_comment, from_domain, policy)
+
 def check_spf(ip, mail_from, helo):
     res, reason = spf.check2(ip, mail_from, helo)
     return SPFAuthenticationResult(result=res, reason=reason, smtp_mailfrom=mail_from, smtp_helo=helo)
@@ -102,124 +313,6 @@ def check_arc(msg, logger=None, dnsfunc=None):
 
 def check_dmarc(msg, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=False):
 
-    def check_psddmarc_list(psdname, dnsfunc=dns_query):
-        """Check psddmarc.org list of PSD DMARC participants"""
-        try:
-            # If the PSD registry is locally available, use it.
-            psdfile_name = resource_filename('authheaders', 'psddmarc.csv')
-            psd_file = open(psdfile_name)
-            psds = []
-            for line in psd_file.readlines():
-                sp = line.split(',')
-                if sp[1] == 'Active':
-                    psds += sp[0]
-            if psdname in psds:
-                return True
-            else:
-                return False
-        except:
-            # If not, use the DNS query list.
-            psd_list_host = '.psddmarc.org'
-            psd_lookup = psdname + psd_list_host
-            answer = dnsfunc(psd_lookup)
-            if answer:
-                return True
-            else:
-                return False
-
-    def dmarc_per_from(from_domain, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=False):
-        # Get dmarc record for domain
-        if(dnsfunc):
-            record, orgdomain = receiver_record(from_domain, dnsfunc=dnsfunc)
-        else:
-            record, orgdomain = receiver_record(from_domain)
-        # Report if DMARC record is From Domain or Org Domain
-        if record and orgdomain:
-            result_comment = 'Used Org Domain Record'
-        elif record:
-            result_comment = 'Used From Domain Record'
-
-        # Get psddmarc record if doing PSD DMARC, no DMARC record, and PSD is
-        #  listed
-        psddomain = False
-        if (not record) and psddmarc:
-            org_domain = get_org_domain(from_domain)
-            if(dnsfunc):
-                if check_psddmarc_list(org_domain.split('.',1)[-1],
-                                       dnsfunc=dnsfunc):
-                    record, _ = receiver_record(org_domain.split('.',1)[-1],
-                                                dnsfunc=dnsfunc)
-            else:
-                if check_psddmarc_list(org_domain.split('.',1)[-1]):
-                    record, _ = receiver_record(org_domain.split('.',1)[-1])
-            if record:
-                psddomain = True
-                result_comment = 'Used Public Suffix Domain Record'
-
-        if record and record.get('p'): # DMARC P tag is mandatory
-            # find policy
-            policy = record['p']
-            if policy[-1:] == '\\':
-                policy = policy[:-1]
-            try:
-                sp = record['sp']
-                if sp[-1:] == '\\':
-                    sp = sp[:-1]
-            except KeyError:
-                sp = policy
-            try:
-                np = record['np']
-                if np[-1:] == '\\':
-                    np = np[:-1]
-            except KeyError:
-                np = None
-
-            if orgdomain or psddomain:
-                if np:
-                    exists = False
-                    for qtype in ['a', 'mx', 'aaaa']:
-                        if(dnsfunc):
-                            res = dnsfunc(from_domain)
-                        else:
-                            res = dns_query(from_domain, qtype)
-                        if res:
-                            exists = True
-                            break
-                    if exists:
-                        policy = sp
-                    else:
-                        policy = np
-                else:
-                    policy = sp
-
-            adkim = record.get('adkim', 'r')
-            aspf  = record.get('aspf',  'r')
-
-            # get result
-            result = "fail"
-            if spf_result and spf_result.result == "pass":
-                # The domain in SPF results often includes the local part, even though
-                # generally it SHOULD NOT (RFC 7601, Section 2.7.2, last paragraph).
-                mail_from_domain = get_domain_part(spf_result.smtp_mailfrom)
-                spf_result.smtp_mailfrom = mail_from_domain
-                if aspf == "s" and from_domain == mail_from_domain:
-                    result = "pass"
-                elif aspf == "r" and get_org_domain(from_domain) == get_org_domain(mail_from_domain):
-                    result = "pass"
-
-            if dkim_result and dkim_result.result == "pass":
-                if adkim == "s" and from_domain == dkim_result.header_d:
-                    result = "pass"
-                elif adkim == "r" and get_org_domain(from_domain) == get_org_domain(dkim_result.header_d):
-                    result = "pass"
-        else:
-            # If no DMARC record, no result
-            result = 'none'
-            result_comment = ''
-            from_domain = ''
-            policy = ''
-        return(result, result_comment, from_domain, policy)
-
     # get from domain
     headers, _ = rfc822_parse(msg)
     from_headers = [a[1] for a in getaddresses(x[1].decode(errors='ignore').strip() for x in headers if x[0].lower() == b"from")]
