@@ -39,7 +39,10 @@ def answer_to_dict(answer):
     '''Turn the DNS DMARC answer into a dict of tag:value pairs.'''
     a = answer.strip('"').strip(' ')
     rawTags = [t.split('=') for t in a.split(';') if t]
-    retval = {t[0].strip(): t[1].strip() for t in rawTags}
+    retval = {t[0].strip().lower(): t[1].strip().lower() for t in rawTags}
+    # Simpler to lowercase everything and put 'v' back.  Already validated
+    # before answer_to_dict is called, so should be fine.
+    retval['v'] = 'DMARC1'
     return retval
 
 def dns_query(name, qtype='TXT'):
@@ -86,7 +89,7 @@ def lookup_receiver_record(host, dnsfunc=dns_query):
             # http://tools.ietf.org/html/draft-kucherawy-dmarc-base-04#section-7.1
             if str(result)[:9] == '"v=DMARC1':
                 if not tags:
-                    tags = answer_to_dict(str(answer[0]))
+                    tags = answer_to_dict(str(result))
                 else:
                     return {} # Multiple DMARC records
         if tags:
